/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.perseus.pool.lib;

import org.objectweb.perseus.pool.api.Pool;
import org.objectweb.perseus.pool.api.PoolAttributes;
import org.objectweb.perseus.pool.api.PoolMatchFactory;
import org.objectweb.perseus.pool.api.PoolException;
import org.objectweb.perseus.dependency.api.DependencyGraph;
import org.objectweb.perseus.dependency.api.DeadLockException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.ArrayList;
import java.util.Calendar;

/**
 * The class <b>LArrayPool</b> implements a Pool as an array of PoolResource,
 * managing free/active resources through two ArrayList objects. An unlimited
 * size and a timeout features are supported. This implementation is thread
 * safe. This implementation is able to manage a TTL for the free pooled resources,
 * but by default the resource has not a TTL (0 value).
 *
 * @author S.Chassande-Barrioz
 */
public class ArrayListPool
		implements Pool, PoolAttributes, BindingController {

	public final static String POOL_MATCH_FACTORY_BINDING = "pool-match-factory";
	public final static String DEPENDENCY_GRAPH_BINDING = "dependency-graph";

    /**
     * The int value used for the timeout or the max size
     */
    public final static int UNLIMITED = -1;

    /**
     * The default minimum size of the pool: 0
     */
    public final static int DEFAULT_MIN_SIZE = 0;

    /**
     * The default maximum size of the pool: UNLIMITED
     */
    public final static int DEFAULT_MAX_SIZE = UNLIMITED;

    /**
     * The default time out for a request: 100ms
     */
    public final static int DEFAULT_TIME_OUT = 300;

    private PoolMatchFactory matchFactory = null;
    private long ttl = 0;
    private long timeOut = DEFAULT_TIME_OUT;
    private int minSize = DEFAULT_MIN_SIZE;
    private int maxSize = DEFAULT_MAX_SIZE;
    private List free = new ArrayList();
    private List timesOfFree = new ArrayList();
    private List used = new ArrayList();
    private List users = new ArrayList();
	private boolean started = false;
	private Logger logger = null;
	private DependencyGraph dg = null;


	/**
	 * destroy the resource that the ttl is expired
	 */
	private void cleanExpiredResource() {
		if (ttl > 0) {
			long time = Calendar.getInstance().getTimeInMillis();
			for(int i=0; i<timesOfFree.size();) {
				long l = ((Long) timesOfFree.get(i)).longValue();
				if ((time - l) > ttl) {
					timesOfFree.remove(i);
					Object resource = free.remove(i);
					if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
					    logger.log(BasicLevel.INFO, "Resource has expired and is destroyed: '" + resource + "'.");
					}
					matchFactory.destroyResource(resource);
				} else {
					i++;
				}
			}
		}
	}

	private void updateToMinsize(Object hints) throws PoolException {
		cleanExpiredResource();
		int delta = minSize - free.size() - used.size();
		if (matchFactory != null && delta>0) {
			Object resource = null;
			for(int i=0; i<delta; i++) {
				resource = matchFactory.createResource(hints);
				if (resource == null) {
					if (started) {
						throw new PoolException("The pool resource factory does not create a new resource with the hints: " + hints);
					} else if (logger != null) {
						logger.log(BasicLevel.WARN, "The pool resource factory does not create a new resource with the hints: " + hints);
					}
				} else {
					if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
					    logger.log(BasicLevel.INFO, "Resource created: '" + resource + "'.");
					}
					free.add(resource);
					if (ttl > 0) {
						timesOfFree.add(new Long(Calendar.getInstance().getTimeInMillis()));
					}
				}
			}
		}
	}

	private void updateToMaxsize() {
		cleanExpiredResource();
		if (maxSize != UNLIMITED) {
			int delta = free.size() + used.size() - maxSize;
			//Remove the free pool resource if availlable
			if (delta > 0) {
				for (int i = 0; i < delta && !free.isEmpty(); i++) {
				    int j = free.size() - 1;
					Object resource = free.remove(j);
					if (ttl > 0) {
					    timesOfFree.remove(j);
					}
					if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
					    logger.log(BasicLevel.INFO, "Resource destroyed: '" + resource + "'.");
					}
					matchFactory.destroyResource(resource);
				}
			}
		}
	}


    // IMPLEMENTATION OF METHODS FROM THE BindingController INTERFACE //
    //----------------------------------------------------------------//

    public String[] listFc() {
        return new String[] {
			POOL_MATCH_FACTORY_BINDING, DEPENDENCY_GRAPH_BINDING };
    }

    public Object lookupFc(String clientItfName) {
        if (POOL_MATCH_FACTORY_BINDING.equals(clientItfName)) {
            return matchFactory;
		} else if (DEPENDENCY_GRAPH_BINDING.equals(clientItfName)) {
            return dg;
		} else {
            return null;
		}
    }

    public void bindFc(String clientItfName, Object serverItf) {
        if (POOL_MATCH_FACTORY_BINDING.equals(clientItfName)) {
            matchFactory = (PoolMatchFactory) serverItf;
            try {
                updateToMinsize(null);
            } catch (Exception e) {
            }
		} else if (DEPENDENCY_GRAPH_BINDING.equals(clientItfName)) {
            dg = (DependencyGraph) serverItf;
        } else if ("logger".equals(clientItfName)) {
			logger = (Logger) serverItf;
		}
    }

    public void unbindFc(String clientItfName) {
        if (POOL_MATCH_FACTORY_BINDING.equals(clientItfName)) {
            matchFactory = null;
		} else if (DEPENDENCY_GRAPH_BINDING.equals(clientItfName)) {
            dg = null;
		}
    }


    // IMPLEMENTATION OF METHODS FROM THE Pool INTERFACE //
    //---------------------------------------------------//

    /**
     * <b>getResource</b> is used to allocate a PoolResource from the Pool.
     * Some hints are passed in order to specialise the matching or creation
     * of PoolResource.
     * @param hints	Some properties to specialise the matching or the creation
     *				of PoolResource.
     * @return		The PoolResource allocated from the Pool.
     */
    public synchronized Object getResource(Object hints) throws PoolException {
		try {
			return getResource(hints, null);
		} catch (DeadLockException e) {
			//not possible because no user is specified
			throw new PoolException(e);
		}
	}
    public synchronized Object getResource(Object hints, Object user)
			throws PoolException, DeadLockException {
		if (!started) {
			started = true;
			updateToMaxsize();
		}
		//System.out.println("getResource: max=" + (maxSize == UNLIMITED ? "U" : "" + maxSize) + ", free=" + free.size() + ", used=" + used.size() +")");
        if (maxSize != UNLIMITED && used.size() >= maxSize) {
            if (timeOut != 0) {
				int vertexIdx = 0;
				try {
					if (user != null && dg != null) {
						int s = users.size();
						for(vertexIdx=0; vertexIdx<s; vertexIdx++) {
							Object dst = users.get(vertexIdx);
							if (dst != null) {
								if (!dg.addVertex(user, dst)) {
									throw new DeadLockException("Impossible to allocate a resource: no resource is availlable and a dead lock occurs if the current thread wait a resource");
								}
							}
						}
					}
					if (timeOut > 0) {
						long to = timeOut;
						while(maxSize != UNLIMITED && used.size() >= maxSize && to > 0) {
							long exectime = System.currentTimeMillis();
							wait(timeOut);
							exectime = System.currentTimeMillis() - exectime;
							to -= exectime;
						}
					} else {
						while (maxSize != UNLIMITED && used.size() >= maxSize) {
							wait();
						}
					}
				} catch (InterruptedException e) {
					throw new PoolException("Pool wait has been interrupted: ", e);
				} finally {
					if (user != null && dg != null) {
						for(int i=0; i<vertexIdx; i++) {
							Object dst = users.get(i);
							if (dst != null) {
								dg.removeVertex(user, dst);
							}
						}
					}
				}
				if (maxSize != UNLIMITED && used.size() >= maxSize) {
                    throw new PoolException(
							"No more resource available in pool, time out exceeded ("
							+ timeOut + ") ! (pool max size=" + maxSize + ")");
                }
			} else {
				throw new PoolException("No more resource available in pool (max="
						+ maxSize + ")");
            }
        }
		updateToMinsize(hints);
		Object resource = null;
        if (free.isEmpty()) {
            resource = matchFactory.createResource(hints);
	        if (resource == null) {
		        throw new PoolException("The pool resource factory does not create a new resource with the hints: " + hints);
	        }
			if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
			    logger.log(BasicLevel.INFO, "Resource created: '" + resource + "'.");
			}
	        used.add(resource);
        }
        int freesize = free.size() -1;
        for (int i = freesize; resource == null && i >=0; i--) {
            resource = free.get(i);
            if (matchFactory.matchResource(resource, hints)) {
                free.remove(i);
				if (ttl > 0) {
                	timesOfFree.remove(i);
				}
                used.add(resource);
				if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
				    logger.log(BasicLevel.INFO, "Resource created: '" + resource + "'.");
				}
			} else {
				resource = null;
			}
        }
	    if (resource == null) {
	        resource = matchFactory.createResource(hints);
		    if (resource == null) {
			    throw new PoolException("The pool resource factory does not create a new resource with the hints: " + hints);
		    }
			if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
			    logger.log(BasicLevel.INFO, "Resource created: '" + resource + "'.");
			}
	        used.add(resource);
	    }
		if (resource != null) {
			if (dg != null) {
				if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
					if (user == null) {
						logger.log(BasicLevel.DEBUG, "The resource '"
								+ resource +"' is fetched.");
					} else {
						logger.log(BasicLevel.DEBUG, "The user '" + user
								+ "' fetchs the resource '" + resource +"'.");
					}
				}
				users.add(user);
			}
			return resource;
		}
        throw new PoolException("No free resource matching to the hints: " + hints);
    }

    public synchronized int getSize() {
        return free.size() + used.size();
    }
    
    public synchronized int getFreeResourceNumber() {
        return free.size();
    }
    public synchronized int getUsedResourceNumber() {
        return used.size();
    }
    public synchronized Collection getUsers() {
        return Collections.unmodifiableCollection(users);
    }
    
    /**
     * <b>releaseResource</b> releases a PoolResource in order to allow the
     * Pool to recycle this PoolResource.
     * @param resource	The PoolResource to be released.
     */
    public synchronized void releaseResource(Object resource) throws PoolException {
        //System.out.println("releaseResource: max=" + (maxSize == UNLIMITED ? "U" : "" + maxSize) + ", free=" + free.size() + ", used=" + used.size() +")");
        boolean removed = false;
		for(int i=(used.size() -1); !removed && i>=0; i--) {
			if (resource == used.get(i)) {
				removed = true;
				used.remove(i);
				if (dg != null) {
					Object user = users.remove(i);
					if (logger != null && logger.isLoggable(BasicLevel.DEBUG)) {
						if (user == null) {
							logger.log(BasicLevel.DEBUG, "The resource '"
									+ resource +"' is released.");
						} else {
							logger.log(BasicLevel.DEBUG, "The user '" + user
									+ "' releases the resource '" + resource +"'.");
						}
					}
				}
			}
		}
		if (!removed) {
			throw new PoolException("Release an unused resource:" + resource);
		}
        if (maxSize != UNLIMITED
				&& (free.size() + used.size()) > maxSize) {
			if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
			    logger.log(BasicLevel.INFO, "Resource destroyed: '" + resource + "'.");
			}
			matchFactory.destroyResource(resource);
        } else {
            free.add(resource);
			if (ttl > 0) {
				timesOfFree.add(new Long(Calendar.getInstance().getTimeInMillis()));
			}
            //signal that a resource has been free
            notify();
        }
    }


    // IMPLEMENTATION OF THE PoolAttributes INTERFACE //
    //------------------------------------------------//

    /**
     * <b>getTimeout</b> retrieves the timeout assigned to this Pool.
     * @return	The timeout currently assigned to this Pool.
     */
    public long getTimeout() {
        return timeOut;
    }

    /**
     * <b>getMinSize</b> retrieves the minimum size assigned to this Pool.
     * @return	The minimum size currently assigned to this Pool.
     */
    public int getMinSize() {
        return minSize;
    }

    /**
     * <b>getMaxSize</b> retrieves the maximum size assigned to this Pool.
     * @return	The maximum size currently assigned to this Pool.
     */
    public int getMaxSize() {
        return maxSize;
    }

    /**
     * <b>setTimeout</b> assigns a timeout to this Pool.
     * @param crto		The timeout to be assigned.
     */
    public void setTimeout(long crto) {
        timeOut = crto;
    }

    /**
     * <b>setMinSize</b> assigns a minimum size to this Pool.
     * @param minsize	The minimum size to be assigned.
     */
    public void setMinSize(int minsize) throws Exception {
        minSize = minsize;
		if (started) {
        	updateToMinsize(null);
		}
    }

    /**
     * <b>setMaxSize</b> assigns a maximum size to this Pool.
     * @param maxsize	The maximum size to be assigned.
     */
    public void setMaxSize(int maxsize) throws Exception {
        maxSize = maxsize;
		if (started) {
        	updateToMaxsize();
		}
    }

	public long getTTL() {
		return ttl;
	}

	public void setTTL(long ttl) {
		this.ttl = ttl;
	}
}
