/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.pool.lib;

import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.perseus.pool.api.Pool;
import org.objectweb.perseus.pool.api.PoolAttributes;
import org.objectweb.perseus.pool.api.PoolMatchFactory;
import org.objectweb.perseus.pool.api.PoolException;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;

/**
 * The class <b>LArrayPool</b> implements a Pool as an array of PoolResource,
 * managing free/active resources through lists of cells encapsulating the
 * resources.
 *
 * @author P.Dechamboux
 */
public class LArrayPool implements Pool, PoolAttributes, BindingController {

	public final static String POOL_MATCH_FACTORY_BINDING = "pool-match-factory";

	private PRCell arrayPool[] = null;
	private int minSize = 0;
	private int maxSize = 0;
  private int newMinSize = 0;
  private int newMaxSize = 0;
	private long timeout = 0;
	private int curSize = 0;
	private int freeSize = 0;
	private PoolMatchFactory matchFactory = null;
	private PRCell freeList = null;
	private PRCell activeList = null;

	// IMPLEMENTATION OF METHODS FROM THE UserBindingController INTERFACE //
	//--------------------------------------------------------------------//

    public String[] listFc() {
        return new String[] { POOL_MATCH_FACTORY_BINDING };
    }

	public Object lookupFc(String clientItfName) {
		if (POOL_MATCH_FACTORY_BINDING.equals(clientItfName))
			return matchFactory;
		else
			return null;
	}

	public void bindFc(String clientItfName, Object serverItf) {
		if (POOL_MATCH_FACTORY_BINDING.equals(clientItfName))
			matchFactory = (PoolMatchFactory) serverItf;
	}

	public void unbindFc(String clientItfName) {
		if (POOL_MATCH_FACTORY_BINDING.equals(clientItfName))
			matchFactory = null;
	}


	// IMPLEMENTATION OF METHODS FROM THE PoolAttributes INTERFACE //
	//-------------------------------------------------------------//

	public long getTimeout() {
		return timeout;
	}

	public int getMinSize() {
		return newMinSize;
	}

	public int getMaxSize() {
		return newMaxSize;
	}

	public synchronized void setTimeout(long crto) {
	}

	public void setMinSize(int minsize) {
    newMinSize = minsize;
  }

  public void setMaxSize(int maxsize) {
    newMaxSize = maxsize;
  }

	public long getTTL() {
		return -1;
	}

	public void setTTL(long ttl) {
	}

  private void updateSize () throws PoolException {
    if (maxSize != newMaxSize) {
      changeMaxSize(newMaxSize);
      maxSize = newMaxSize;
    }
    if (minSize != newMinSize) {
      changeMinSize(newMinSize);
      minSize = newMinSize;
    }
  }

  private void changeMaxSize(int maxsize) throws PoolException {
		if ((maxsize < 0) || (maxsize < newMinSize) || (maxsize == maxSize))
			return;
		PRCell nap[];
		if (maxsize > maxSize) {
			nap = new PRCell[maxsize];
			int i;
			for (i = 0; i < maxSize; i++)
				nap[i] = arrayPool[i];
			for (; i < maxsize; i++)
				nap[i] = null;
			maxSize = maxsize;
		} else {
			int cursize = curSize;
			if (maxsize < curSize) {
				// Removes as many free cells as possible
				int nbRemove = maxSize - maxsize;
				PRCell remcell;
				while ((freeList != null) && (nbRemove > 0)) {
					remcell = freeList;
					freeList = freeList.remove(freeList);
					matchFactory.destroyResource(remcell.getResource());
					arrayPool[remcell.getIndex()] = null;
					nbRemove--;
					cursize--;
				}
				// Adds non-removed cells to the new maxsize
				nap = new PRCell[cursize];
				maxSize = cursize;
			} else {
				nap = new PRCell[maxsize];
				maxSize = maxsize;
			}
			// Copies to new Pool array
			for (int i = 0, j = 0; i < curSize; i++) {
				if (arrayPool[i] != null)
					nap[j++] = arrayPool[i];
			}
			curSize = cursize;
		}
		arrayPool = nap;
	}

  private void changeMinSize(int minsize) throws PoolException {
		if ((minsize < 0) || (minsize > maxSize) || (minsize == minSize))
			return;
		if (minsize < minSize) {
			minSize = minsize;
			return;
		}
		PRCell newprc;
		while (minsize > curSize) {
			newprc = new PRCell(matchFactory.createResource(null), curSize);
			arrayPool[curSize++] = newprc;
			freeList = newprc.insert(freeList);
			freeSize++;
		}
		minSize = minsize;
	}

	// IMPLEMENTATION OF METHODS FROM THE Pool INTERFACE //
	//---------------------------------------------------//

	public Object getResource(Object hints)
		throws PoolException {
		return getResource(hints, null);
	}
	
	public synchronized Object getResource(Object hints,
			Object user) throws PoolException {
	    updateSize();
		PRCell res = null;
		if (freeList != null) {
			Iterator it = freeList.iterator();
			while (it.hasNext()) {
				res = (PRCell) it.next();
				if (matchFactory.matchResource(res.getResource(), hints)) {
					freeList = res.remove(freeList);
					activeList = res.insert(activeList);
					freeSize--;
					break;
				}
			}
		}
		if (res == null) {
			if (curSize < maxSize) {
				res = new PRCell(matchFactory.createResource(hints), curSize);
				arrayPool[curSize++] = res;
				activeList = res.insert(activeList);
			} else {
				throw new PoolException("No more resource available in pool!!");
			}
		}
// 		printLists();
		return res.getResource();
	}

	public synchronized int getSize() {
		return curSize;
	}

	public synchronized int getFreeResourceNumber() {
        return freeSize;
    }
    public synchronized int getUsedResourceNumber() {
        return curSize - freeSize;
    }
    public synchronized Collection getUsers() {
        return Collections.EMPTY_LIST;
    }
    

	public synchronized void releaseResource(Object resource)
		throws PoolException {
	    updateSize();
		if (activeList == null)
			throw new PoolException("Releases inactive resource - 1!!");
		PRCell prc = activeList.search(resource);
		if (prc == null)
			throw new PoolException("Releases inactive resource - 2!!");
		activeList = prc.remove(activeList);
		freeList = prc.insert(freeList);
		freeSize++;
// 		printLists();
	}


	private void printLists() {
		if (maxSize != 200)
			return;
		System.err.println("minSize=" + minSize + ", maxSize=" + maxSize
			+ ", curSize=" + curSize + ", freeSize=" + freeSize);
		System.err.print("activeList:");
		if (activeList == null)
			System.err.print(" null");
		else {
			PRCell cur = activeList;
			do {
				System.err.print(" " + cur);
				cur = cur.getNext();
			} while (cur != activeList);
		}
		System.err.println(".");
		System.err.print("freeList:");
		if (freeList == null)
			System.err.print(" null");
		else {
			PRCell cur = freeList;
			do {
				System.err.print(" " + cur);
				cur = cur.getNext();
			} while (cur != freeList);
		}
		System.err.println(".");
	}

}

class PRCell {
	private PRCell next;
	private PRCell prev;
	private Object poolResource;
	private int index;

	PRCell(Object pr, int ind) {
		next = prev = null;
		poolResource = pr;
		index = ind;
	}

	int getIndex() {
		return index;
	}

	PRCell getNext() {
		return next;
	}

	Object getResource() {
		return poolResource;
	}

	void destroy() throws PoolException {
	}

	PRCell insert(PRCell prclist) {
		if (prclist == null) {
			next = prev = this;
		} else {
			PRCell prev2 = prclist.prev;
			prclist.prev.next = this;
			prclist.prev = this;
			next = prclist;
			prev = prev2;
		}
		return this;
	}

	Iterator iterator() {
		return new PRCIter(this);
	}

	PRCell remove(PRCell prclist) throws PoolException {
		if (prclist == this) {
			if (prev == this) {
				next = prev = null;
				return null;
			}
			PRCell next2 = prclist.next;
			prclist.prev.next = next2;
			next2.prev = prclist.prev;
			next = prev = null;
			return next2;
		}
		PRCell rem = prclist.next;
		while ((rem != this) && (rem != prclist))
			rem = rem.next;
		if (rem == prclist)
			throw new PoolException("LArrayPool.PRCell.remove: cell not on this list!!");
		rem.next.prev = rem.prev;
		rem.prev.next = rem.next;
		next = prev = null;
		return prclist;
	}

	PRCell search(Object pr) {
		PRCell end = this;
		PRCell cur = this;
		do {
			if (cur.poolResource == pr)
				return cur;
			cur = cur.next;
		} while (cur != end);
		return null;
	}

	void setIndex(int ind) {
		index = ind;
	}

	class PRCIter implements Iterator {
		PRCell cur, end;

		PRCIter(PRCell list) {
			if (list != null)
				end = list.prev;
			cur = list;
		}

		public boolean hasNext() {
			return cur != null;
		}

		public Object next() {
			if (cur == null)
				return null;
			if (cur == end) {
				cur = null;
				return end;
			}
			cur = cur.next;
			return cur.prev;
		}

		public void remove() {
		}
	}
}